@@ -10,7 +10,7 @@ from .tenant import Tenant
 from . import __CONNECTOR_PYCAROL__
 from . import __version__
 from . organization import Organization
-
+from .exceptions import CarolApiResponseException, InvalidToken
 
 class Carol:
     """
@@ -32,7 +32,7 @@ class Carol:
         verbose: `bool` , default `False`.
             If True will print the header, method and URL of each API call.
         organization: `str` , default `None`.
-            Organization domain. 
+            Organization domain.
         environment: `str`, default `carol.ai`,
             Which Carol's environment to use. There are three possible values today.
 
@@ -67,31 +67,27 @@ class Carol:
     def __init__(self, domain=None, app_name=None, auth=None, connector_id=None, port=443, verbose=False,
                  organization=None, environment=None, host=None):
 
-        settings = dict()
-        if auth is None and domain is None:
-
-            domain = os.getenv('CAROLTENANT')
-            app_name = os.getenv('CAROLAPPNAME')
-
-            assert domain and app_name, \
-                f"One of the following env variables are missing:\n CAROLTENANT: {domain}\n CAROLAPPNAME: {app_name}"
-
+        if auth is None:
             carol_user = os.getenv('CAROLUSER')
             carol_pw = os.getenv('CAROLPWD')
 
             if carol_user and carol_pw:
                 auth = PwdAuth(user=carol_user, password=carol_pw)
-
             else:
                 auth_token = os.getenv('CAROLAPPOAUTH')
                 connector_id = os.getenv('CAROLCONNECTORID')
-
-                assert domain and app_name and auth_token and connector_id,\
-                    "One of the following env variables are missing:\n " \
-                    f"CAROLTENANT: {domain}\nCAROLAPPNAME: {app_name}" \
-                    f"\nCAROLAPPOAUTH: {auth}\nCAROLCONNECTORID: {connector_id}\n"
-
                 auth = ApiKeyAuth(auth_token)
+            if auth is None:
+                raise ValueError("either `auth` method or pycarol env variables must be set.")
+
+        if domain is None:
+            domain = os.getenv('CAROLTENANT')
+            if domain is None:
+                raise ValueError("`domain` must be set.")
+
+        if app_name is None:
+            app_name = os.getenv('CAROLAPPNAME', ' ')
+
 
         if connector_id is None:
             if auth.connector_id is None:
@@ -113,9 +109,10 @@ class Carol:
         self.app_name = app_name
         self.port = port
         self.verbose = verbose
+        self._host_string = host
         self.host = self._set_host(domain=self.domain, organization=self.organization,
                                    environment=self.environment, host=host)
-        self.tenant = Tenant(self).get_tenant_by_domain(domain)
+        self._tenant = None
         self.connector_id = connector_id
         self.auth = auth
         self.auth.set_connector_id(self.connector_id)
@@ -125,6 +122,14 @@ class Carol:
         self.org = None
 
 
+
+    @property
+    def tenant(self):
+        if self._tenant is None:
+            self._tenant = Tenant(self).get_tenant_by_domain(self.domain)
+        return self._tenant
+
+
     @staticmethod
     def _set_host(domain, organization, environment, host):
         """
@@ -201,8 +206,8 @@ class Carol:
 
     def call_api(self, path, method=None, data=None, auth=True, params=None, content_type='application/json', retries=8,
                  session=None, backoff_factor=0.5, status_forcelist=(502, 503, 504, 524), downloadable=False,
-                 method_whitelist=frozenset(['HEAD', 'TRACE', 'GET', 'PUT', 'OPTIONS', 'DELETE']), errors='raise',
-                 extra_headers=None,
+                 method_whitelist=frozenset(['HEAD', 'TRACE', 'GET', 'PUT', 'OPTIONS', 'DELETE', 'POST']), errors='raise',
+                 extra_headers=None, files=None,
                  **kwds):
         """
         This method handles all the API calls.
@@ -242,6 +247,8 @@ class Carol:
                 then invalid request will return the request response
             extra_headers: `dict` default `None`
                 extra headers to be sent.
+            files: `dict` default `None`
+                Used when uploading files to carol. This will be sent to :class: `requests.request`
             kwds: `dict` default `None`
                 Extra parameters to be sent to :class: `requests.request`
 
@@ -272,7 +279,8 @@ class Carol:
             pass
 
         elif (method == 'POST') or (method == 'DELETE') or (method == 'PUT'):
-            headers['content-type'] = content_type
+            if content_type is not None:
+                headers['content-type'] = content_type
 
             if content_type == 'application/json':
                 data_json = data
@@ -288,7 +296,7 @@ class Carol:
                                               status_forcelist=status_forcelist, method_whitelist=method_whitelist)
 
             response = session.request(method=method, url=url, data=data, json=data_json,
-                                       headers=headers, params=params, **kwds)
+                                       headers=headers, params=params, files=files, **kwds)
 
             if self.verbose:
                 if data_json is not None:
@@ -307,15 +315,17 @@ class Carol:
                     return {}
                 return json.loads(response.text)
 
-            elif (response.reason == 'Unauthorized') and isinstance(self.auth,PwdAuth):
+            elif (response.reason == 'Unauthorized') and isinstance(self.auth, PwdAuth):
                 if response.json().get('possibleResponsibleField') in ['password', 'userLogin']:
-                    raise Exception(response.text)
+                    raise InvalidToken(response.text)
                 self.auth.get_access_token()  #It will refresh token if Unauthorized
-                __count+=1
-                if __count<5: #To avoid infinity loops
+                __count += 1
+                if __count < 5: #To avoid infinity loops
                     continue
                 else:
                     raise Exception('Too many retries to refresh token.\n', response.text, response.status_code)
+            elif (response.status_code == 404):
+                raise CarolApiResponseException(response.text, response.status_code)
 
             raise Exception(response.text, response.status_code)
 
@@ -395,9 +405,15 @@ class Carol:
         else:
             raise Exception("Auth object not set. Can't fetch token.")
 
-    def switch_environment(self, env_name=None, env_id=None, app_name=None):
+    def switch_org_level(self):
+
+        org = self._current_org()
+        self.auth.switch_org_context(org['mdmId'])
+
+    def switch_environment(self, env_name=None, env_id=None, app_name=None, org_name=None, org_id=None):
         """
-        Switch environments. If the user has access to this environment, it will be "logged in" in this new environment.
+        Switch org/environments. If the user has access to this environment, it will be "logged in" in this new
+        org/environment.
 
         Args:
 
@@ -408,6 +424,10 @@ class Carol:
             app_name: `str` default `None`
                 App name in the target environment to switch the context to.
                 Only needed with using CDS.
+            org_name: `str` default `None`
+                The organization name to switch context to. If the same keep it `None`
+            org_id: `str` default `None`
+                The organization id to switch context to. If the same keep it `None`
 
         Returns:
             self
@@ -428,6 +448,16 @@ class Carol:
         if self.org is None:
             self.org = Organization(self).get_organization_info(self.organization)
 
+        current = self.get_current()
+
+        if (org_id is not None and org_id != current['org_id']) or (org_name is not None and org_name != current['org_name'] ):
+            # Switch to org context.
+            self.switch_org_level()
+            if org_id is None:
+                org_id = Organization(self).get_organization_info(org_name)['mdmId']
+            # Switch org.
+            self.auth.switch_org_context(org_id)
+
         if env_name:
             env_id = Tenant(self).get_tenant_by_domain(env_name)['mdmId']
         elif env_id is None:
@@ -436,12 +466,16 @@ class Carol:
         self.auth.switch_context(env_id=env_id)
 
         self.domain = env_name
-        self.app_name = app_name #TODO: Today we cannot use CDS without a valid app name.
-        self.tenant = Tenant(self).get_tenant_by_domain(env_name)
+        self.app_name = app_name # TODO: Today we cannot use CDS without a valid app name.
+        self._tenant = self._current_env()
+        self.organization = self._current_org()['mdmName']
+
+        self.host = self._set_host(domain=self.domain, organization=self.organization,
+                                   environment=self.environment, host=self._host_string)
 
         return self
 
-    def switch_context(self, env_name=None, env_id=None, app_name=None):
+    def switch_context(self, env_name=None, env_id=None, app_name=None, org_name=None, org_id=None):
         """
         Context manager to temporary have access to a second environment
 
@@ -481,19 +515,65 @@ class Carol:
 
         class SwitchContext(object):
 
-            def __init__(self, parent_context, env_name=None, env_id=None, app_name=None):
+            def __init__(self, parent_context, env_name=None, env_id=None, org_id=None, org_name=None, app_name=None):
                 self.parent_context = parent_context
                 self.env_name = env_name
                 self.env_id = env_id
                 self.app_name = app_name
+                self.org_id = org_id
+                self.org_name = org_name
 
             def __enter__(self):
-
-                self.parent_context.switch_environment(env_name= self.env_name, env_id=self.env_id, app_name=self.app_name)
-
+                self.parent_context.switch_environment(env_name=self.env_name, env_id=self.env_id,
+                                                       app_name=self.app_name, org_name=self.org_name,
+                                                       org_id=self.org_id)
                 return self.parent_context
 
             def __exit__(self, exc_type, exc_val, exc_tb):
                 del self.parent_context
 
-        return SwitchContext(parent_context=copy.deepcopy(self), env_name=env_name, env_id=env_id, app_name=app_name)
+        return SwitchContext(parent_context=copy.deepcopy(self), env_name=env_name, env_id=env_id, app_name=app_name,
+                             org_name=org_name, org_id=org_id,)
+
+
+    def _current_env(self):
+        return self.call_api('v1/tenants/current')
+
+    def _current_org(self):
+        return self.call_api('v1/organizations/current')
+
+    def get_current(self, level='all'):
+        """
+        Get current org/env information.
+
+        Args:
+
+            level: `str`:
+                Possible Values:
+                    "all": To get organization and environment information.
+                    "org": To get organization information.
+                    "env": To get environment information.
+
+        Returns: `dict`
+            Dictionary with keys org_Id, org_name, env_id, env_name
+
+        """
+
+        env = {}
+        org = {}
+
+        if level.lower() not in ['org', 'env', 'all']:
+            raise ValueError(f"level should be 'all', 'org', 'env', {level} was passed.")
+
+        if level.lower() == 'env' or level.lower() == 'all':
+            env = self._current_env()
+
+        if level.lower() == 'org' or level.lower() == 'all':
+            org = self._current_org()
+
+        return {
+            "env_name": env.get('mdmName'),
+            "env_id": env.get('mdmId'),
+            "org_name": org.get('mdmName'),
+            "org_id": org.get('mdmId'),
+        }
